#include <glib.h>
#include <glib-object.h>
#include "makuobject.h"
#include "marshallers.h"


static void maku_object_class_init (gpointer g_class, gpointer g_class_data);
static void maku_object_init (GTypeInstance *instance, gpointer g_class);
static GObject * maku_object_constructor (GType type, guint n_construct_prop, GObjectConstructParam *construct_prop);
static void maku_object_finalize (GObject *object);
static DBusGProxy * get_dbus_gproxy (MakuObject *self);
void name_owner_changed_callback (DBusGProxy *gproxy, gchar *arg1, gchar *arg2, gchar *arg3, gpointer user_data);



static gpointer maku_object_parent_class = NULL;
static guint maku_object_signal [LAST_SIGNAL] = {0};


#define DBUS_NAME		"org.freedesktop.DBus"
#define DBUS_PATH		"/org/freedesktop/DBus"
#define DBUS_INTERFACE		"org.freedesktop.DBus"


// returns the type
GType maku_object_get_type (void)
{
	static GType type = 0;
	
	if (G_UNLIKELY (type == 0))
	{
		static const GTypeInfo info = 
		{
			sizeof (MakuObjectClass),
			NULL,
			NULL,
			(GClassInitFunc) maku_object_class_init,
			NULL,
			NULL,
			sizeof (MakuObject),
			0,
			(GInstanceInitFunc) maku_object_init,
			NULL
		};
		
		type = g_type_register_static (G_TYPE_OBJECT, "MakuObject", &info, 0);
	}
	return type;
}



// class init
static void maku_object_class_init (gpointer g_class, gpointer g_class_data)
{
	GObjectClass *gobj_class = G_OBJECT_CLASS (g_class);
	MakuObjectClass *maku_obj_class = MAKU_OBJECT_CLASS (g_class);
	
	maku_object_parent_class = g_type_class_peek_parent (maku_obj_class);
	
	gobj_class->constructor = maku_object_constructor;
	gobj_class->finalize = maku_object_finalize;
	
	/* signals */
	/* We are not concerned with the args that NameOwnerChanged carries... */
	maku_object_signal [NAME_OWNER_CHANGED] = g_signal_new ("name-owner-changed",\
						 TYPE_MAKU_OBJECT,\
						 G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,\
						 0,\
						 NULL,\
						 NULL,\
						 g_cclosure_marshal_VOID__VOID,\
						 G_TYPE_NONE,\
						 0);
}



// instance init for MakuObject
static void maku_object_init (GTypeInstance *instance, gpointer g_class)
{
	MakuObject *self = (MakuObject *)instance;
	
	// initialize the DBus proxy
	self->gproxy = get_dbus_gproxy (self);
	
	// other stuff!
	self->age = 100;
}



// constructor to override GObject's constructor.
static GObject * maku_object_constructor (GType type, guint n_construct_prop, GObjectConstructParam *construct_prop)
{
	GObject *gobject = NULL;
	MakuObjectClass *maku_obj_class = NULL;
	GObjectClass *gobj_class = NULL;

	maku_obj_class = MAKU_OBJECT_CLASS (g_type_class_peek (TYPE_MAKU_OBJECT));
	gobj_class = G_OBJECT_CLASS (g_type_class_peek_parent (maku_obj_class));
	
	gobject = gobj_class->constructor (type, n_construct_prop, construct_prop);
	
	return gobject;
}



static void maku_object_finalize (GObject *object)
{
	MakuObject *self = MAKU_OBJECT (object);
	
	g_print ("Finalizing object!\n");
	if (self)
		g_object_unref (self->gproxy);
}



/* Creates a DBusGProxy for the object & connects to any interested signals */
static DBusGProxy * get_dbus_gproxy (MakuObject *self)
{
	GError *error = NULL;
	DBusGProxy *gproxy = NULL;
	DBusGConnection *connection = NULL;

	connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (!connection) {
		g_print ("Failed to get connection");
		g_error_free (error);
		return NULL;
	}

	g_assert ((gproxy = dbus_g_proxy_new_for_name (connection,\
							DBUS_NAME,\
							DBUS_PATH,\
							DBUS_INTERFACE)));
							
	
	/* TODO - Attaching signals should be moved to a new function - maybe a wrapper... */
	dbus_g_object_register_marshaller (g_cclosure_user_marshal_VOID__STRING_STRING_STRING, G_TYPE_NONE,\
					G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);
							
	/* Done once for each signal in object lifetime - for sanity purposes */
	dbus_g_proxy_add_signal (gproxy, "NameOwnerChanged", G_TYPE_STRING, G_TYPE_STRING,\
				 G_TYPE_STRING, G_TYPE_INVALID);
	
	/* done for each signal */
	dbus_g_proxy_connect_signal (gproxy, "NameOwnerChanged", G_CALLBACK (name_owner_changed_callback),
			       self, NULL);
			       
	g_print ("Signals connected...\n");
	
	return gproxy;
}



/* Wrapping NameOwnerChanged signal */
void name_owner_changed_callback (DBusGProxy *gproxy, gchar *arg1, gchar *arg2, gchar *arg3, gpointer user_data)
{
	MakuObject *maku_object = MAKU_OBJECT (user_data);
	g_print ("Got a Signal!\nRe-emitting name-owner-changed - as a g-signal!\n");
	g_print ("Signal's Args : [%s] [%s] [%s]\n", arg1, arg2, arg3);
	
	g_signal_emit_by_name (maku_object, "name-owner-changed", NULL);
}











